/*
 * Copyright 2003-2011 Optimatika (www.optimatika.se)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.ojalgo.finance.portfolio;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;

import org.ojalgo.access.Access2D;
import org.ojalgo.matrix.BasicMatrix;
import org.ojalgo.matrix.MatrixBuilder;
import org.ojalgo.matrix.PrimitiveMatrix;
import org.ojalgo.matrix.store.ZeroStore;
import org.ojalgo.random.process.GeometricBrownianMotion;
import org.ojalgo.random.process.MultidimensionalProcess;
import org.ojalgo.random.process.RandomProcess;

public final class SimplePortfolio extends FinancePortfolio {

    private static List<SimpleAsset> toSimpleAssets(final Number[] someWeights) {

        final ArrayList<SimpleAsset> retVal = new ArrayList<SimpleAsset>(someWeights.length);

        for (int i = 0; i < someWeights.length; i++) {
            retVal.add(new SimpleAsset(someWeights[i]));
        }

        return retVal;
    }

    private final List<SimpleAsset> myComponents;
    private final Access2D<?> myCorrelations;
    private transient BasicMatrix myCovarianceMatrix = null;
    private transient Number myMeanReturn;
    private transient BasicMatrix myReturnsVector = null;
    private transient Number myReturnVariance;
    private transient List<BigDecimal> myWeights;
    private transient BasicMatrix myWeightsVector = null;

    public SimplePortfolio(final Access2D<?> aCorrelationsMatrix, final List<SimpleAsset> someAssets) {

        super();

        if ((someAssets.size() != aCorrelationsMatrix.getRowDim()) || (someAssets.size() != aCorrelationsMatrix.getColDim())) {
            throw new IllegalArgumentException("Input dimensions don't match!");
        }

        myCorrelations = aCorrelationsMatrix;
        myComponents = someAssets;
    }

    public SimplePortfolio(final List<SimpleAsset> someAssets) {
        this(ZeroStore.makePrimitive(someAssets.size(), someAssets.size()), someAssets);
    }

    public SimplePortfolio(final Number... someWeights) {
        this(SimplePortfolio.toSimpleAssets(someWeights));
    }

    @SuppressWarnings("unused")
    private SimplePortfolio() {
        this(null, null);
    }

    public RandomProcess<?> forecast(final boolean multidimensional) {

        if (multidimensional) {

            final List<GeometricBrownianMotion> tmpAssetProcesses = new ArrayList<GeometricBrownianMotion>(myComponents.size());

            for (final SimpleAsset tmpAsset : myComponents) {
                tmpAssetProcesses.add(tmpAsset.forecast());
            }

            return new MultidimensionalProcess(tmpAssetProcesses, myCorrelations);

        } else {

            return this.forecast();
        }
    }

    public double getCorrelation(final int aRow, final int aCol) {
        return myCorrelations.doubleValue(aRow, aCol);
    }

    public double getCovariance(final int aRow, final int aCol) {

        final double tmpRowRisk = this.getVolatility(aRow);
        final double tmpColRisk = this.getVolatility(aCol);

        final double tmpCorrelation = this.getCorrelation(aRow, aCol);

        return tmpRowRisk * tmpCorrelation * tmpColRisk;
    }

    @Override
    public double getMeanReturn() {

        if (myMeanReturn == null) {
            final BasicMatrix tmpWeightsVector = this.getWeightsVector();
            final BasicMatrix tmpReturnsVector = this.getReturnsVector();
            myMeanReturn = MarketEquilibrium.calculatePortfolioReturn(tmpWeightsVector, tmpReturnsVector).getNumber();
        }

        return myMeanReturn.doubleValue();
    }

    public double getMeanReturn(final int index) {
        return myComponents.get(index).getMeanReturn();
    }

    @Override
    public double getReturnVariance() {

        if (myReturnVariance == null) {
            final MarketEquilibrium tmpMarketEquilibrium = this.toMarketEquilibrium(BOOLEAN_FALSE);
            final BasicMatrix tmpWeightsVector = this.getWeightsVector();
            myReturnVariance = tmpMarketEquilibrium.calculatePortfolioVariance(tmpWeightsVector).getNumber();
        }

        return myReturnVariance.doubleValue();
    }

    public double getReturnVariance(final int index) {
        return myComponents.get(index).getReturnVariance();
    }

    public double getVolatility(final int index) {
        return myComponents.get(index).getVolatility();
    }

    public BigDecimal getWeight(final int index) {
        return myComponents.get(index).getWeight();
    }

    @Override
    public List<BigDecimal> getWeights() {

        if (myWeights == null) {

            myWeights = new ArrayList<BigDecimal>(myComponents.size());

            for (final SimpleAsset tmpAsset : myComponents) {
                myWeights.add(tmpAsset.getWeight());
            }
        }

        return myWeights;
    }

    public MarketEquilibrium toMarketEquilibrium(final boolean calibrated) {

        final BasicMatrix tmpCovarianceMatrix = this.getCovarianceMatrix();
        final MarketEquilibrium retVal = new MarketEquilibrium(tmpCovarianceMatrix);

        if (calibrated) {
            final BasicMatrix tmpWeightsVector = this.getWeightsVector();
            final BasicMatrix tmpReturnsVector = this.getReturnsVector();
            retVal.calibrate(tmpWeightsVector, tmpReturnsVector);
        }

        return retVal;
    }

    @Override
    protected void reset() {

        myMeanReturn = null;
        myReturnVariance = null;
        myWeights = null;

        myCovarianceMatrix = null;
        myReturnsVector = null;
        myWeightsVector = null;

        for (final SimpleAsset tmpAsset : myComponents) {
            tmpAsset.reset();
        }
    }

    BasicMatrix getCovarianceMatrix() {

        if (myCovarianceMatrix == null) {

            final int tmpSize = myComponents.size();

            final MatrixBuilder<Double> tmpCovaris = PrimitiveMatrix.FACTORY.getBuilder(tmpSize, tmpSize);

            for (int j = 0; j < tmpSize; j++) {
                for (int i = 0; i < tmpSize; i++) {
                    tmpCovaris.set(i, j, this.getCovariance(i, j));
                }
            }

            myCovarianceMatrix = tmpCovaris.build();
        }

        return myCovarianceMatrix;
    }

    BasicMatrix getReturnsVector() {

        if (myReturnsVector == null) {

            final int tmpSize = myComponents.size();

            final MatrixBuilder<Double> tmpReturns = PrimitiveMatrix.FACTORY.getBuilder(tmpSize, 1);

            for (int i = 0; i < tmpSize; i++) {
                tmpReturns.set(i, 0, this.getMeanReturn(i));
            }

            myReturnsVector = tmpReturns.build();
        }

        return myReturnsVector;
    }

    BasicMatrix getWeightsVector() {

        if (myWeightsVector == null) {

            final int tmpSize = myComponents.size();

            final MatrixBuilder<Double> tmpWeights = PrimitiveMatrix.FACTORY.getBuilder(tmpSize, 1);

            for (int i = 0; i < tmpSize; i++) {
                tmpWeights.set(i, 0, this.getWeight(i));
            }

            myWeightsVector = tmpWeights.build();
        }

        return myWeightsVector;
    }

}
