/*
 * Copyright 2003-2011 Optimatika (www.optimatika.se)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.ojalgo.matrix.decomposition;

import java.math.BigDecimal;

import org.ojalgo.access.Access2D;
import org.ojalgo.array.Array1D;
import org.ojalgo.array.Array2D;
import org.ojalgo.constant.PrimitiveMath;
import org.ojalgo.matrix.MatrixUtils;
import org.ojalgo.matrix.store.BigDenseStore;
import org.ojalgo.matrix.store.ComplexDenseStore;
import org.ojalgo.matrix.store.MatrixStore;
import org.ojalgo.matrix.store.PhysicalStore;
import org.ojalgo.matrix.store.PrimitiveDenseStore;
import org.ojalgo.matrix.transformation.Householder;
import org.ojalgo.scalar.ComplexNumber;
import org.ojalgo.type.TypeUtils;
import org.ojalgo.type.context.NumberContext;

/**
 * You create instances of (some subclass of) this class by calling one
 * of the static factory methods: {@linkplain #makeBig()},
 * {@linkplain #makeComplex()} or {@linkplain #makePrimitive()}.
 *
 * @author apete
 */
public abstract class BidiagonalDecomposition<N extends Number> extends InPlaceDecomposition<N> implements Bidiagonal<N> {

    static final class Big extends BidiagonalDecomposition<BigDecimal> {

        Big() {
            super(BigDenseStore.FACTORY);
        }

    }

    static final class Complex extends BidiagonalDecomposition<ComplexNumber> {

        Complex() {
            super(ComplexDenseStore.FACTORY);
        }

    }

    static final class Primitive extends BidiagonalDecomposition<Double> {

        Primitive() {
            super(PrimitiveDenseStore.FACTORY);
        }

    }

    private static final long PRIMITIVE_ELEMENT = 112L; // 4 x 8L

    @SuppressWarnings("unchecked")
    public static final <N extends Number> Bidiagonal<N> make(final Access2D<N> aTypical) {

        final Number tmpNumber = aTypical.get(INT_ZERO, INT_ZERO);

        if (tmpNumber instanceof BigDecimal) {
            return (Bidiagonal<N>) BidiagonalDecomposition.makeBig();
        } else if (tmpNumber instanceof ComplexNumber) {
            return (Bidiagonal<N>) BidiagonalDecomposition.makeComplex();
        } else if (tmpNumber instanceof Double) {
            return (Bidiagonal<N>) BidiagonalDecomposition.makePrimitive();
        } else {
            throw new IllegalArgumentException();
        }
    }

    public static final Bidiagonal<BigDecimal> makeBig() {
        return new BidiagonalDecomposition.Big();
    }

    public static final Bidiagonal<ComplexNumber> makeComplex() {
        return new BidiagonalDecomposition.Complex();
    }

    public static final Bidiagonal<Double> makePrimitive() {
        return new BidiagonalDecomposition.Primitive();
    }

    private transient DiagonalAccess<N> myDiagonalAccessD;
    private transient DecompositionStore<N> myQ1;
    private transient DecompositionStore<N> myQ2;

    protected BidiagonalDecomposition(final DecompositionStore.Factory<N, ? extends DecompositionStore<N>> aFactory) {
        super(aFactory);
    }

    public final boolean compute(final Access2D<?> aStore) {

        this.reset();

        final DecompositionStore<N> tmpStore = this.setInPlace(aStore);

        final int tmpRowDim = this.getRowDim();
        final int tmpColDim = this.getColDim();

        final int tmpLimit = Math.min(tmpRowDim, tmpColDim);

        final Householder<N> tmpHouseholderCol = this.makeHouseholder(tmpRowDim);
        final Householder<N> tmpHouseholderRow = this.makeHouseholder(tmpColDim);

        if (this.isAspectRatioNormal()) {

            for (int ij = INT_ZERO; ij < tmpLimit; ij++) {

                if (((ij + INT_ONE) < tmpRowDim) && tmpStore.generateApplyAndCopyHouseholderColumn(ij, ij, tmpHouseholderCol)) {
                    tmpStore.transformLeft(tmpHouseholderCol, ij + INT_ONE);
                }

                if (((ij + INT_TWO) < tmpColDim) && tmpStore.generateApplyAndCopyHouseholderRow(ij, ij + INT_ONE, tmpHouseholderRow)) {
                    tmpStore.transformRight(tmpHouseholderRow, ij + INT_ONE);
                }
            }

        } else {

            for (int ij = INT_ZERO; ij < tmpLimit; ij++) {

                if (((ij + INT_ONE) < tmpColDim) && tmpStore.generateApplyAndCopyHouseholderRow(ij, ij, tmpHouseholderRow)) {
                    tmpStore.transformRight(tmpHouseholderRow, ij + INT_ONE);
                }

                if (((ij + INT_TWO) < tmpRowDim) && tmpStore.generateApplyAndCopyHouseholderColumn(ij + INT_ONE, ij, tmpHouseholderCol)) {
                    tmpStore.transformLeft(tmpHouseholderCol, ij + INT_ONE);
                }
            }
        }

        return this.computed(BOOLEAN_TRUE);
    }

    public final boolean equals(final MatrixStore<N> aStore, final NumberContext aCntxt) {
        return MatrixUtils.equals(aStore, this, aCntxt);
    }

    public final MatrixStore<N> getD() {
        return this.getInPlace().builder().bidiagonal(this.isAspectRatioNormal(), BOOLEAN_FALSE).build();
    }

    public final MatrixStore<N> getQ1() {
        if (myQ1 == null) {
            myQ1 = this.makeQ1(this.makeEye(this.getRowDim(), this.getMinDim()), BOOLEAN_TRUE);
        }
        return myQ1;
    }

    public final MatrixStore<N> getQ2() {
        if (myQ2 == null) {
            myQ2 = this.makeQ2(this.makeEye(this.getColDim(), this.getMinDim()), BOOLEAN_TRUE);
        }
        return myQ2;
    }

    public final boolean isFullSize() {
        return BOOLEAN_FALSE;
    }

    public final boolean isSolvable() {
        return BOOLEAN_FALSE;
    }

    public final boolean isUpper() {
        return this.isAspectRatioNormal();
    }

    public final MatrixStore<N> reconstruct() {
        return MatrixUtils.reconstruct(this);
    }

    @Override
    public final void reset() {

        super.reset();

        myQ1 = null;
        myQ2 = null;
        myDiagonalAccessD = null;
    }

    @Override
    public final MatrixStore<N> solve(final MatrixStore<N> aRHS) {
        throw new UnsupportedOperationException();
    }

    private final DiagonalAccess<N> makeDiagonalAccessD() {

        final Array2D<N> tmpArray2D = this.getInPlace().asArray2D();

        final Array1D<N> tmpMain = tmpArray2D.sliceDiagonal(0, 0);
        Array1D<N> tmpSuper;
        Array1D<N> tmpSub;

        if (this.isAspectRatioNormal()) {
            tmpSuper = tmpArray2D.sliceDiagonal(0, 1);
            tmpSub = null;
        } else {
            tmpSub = tmpArray2D.sliceDiagonal(1, 0);
            tmpSuper = null;
        }

        return new DiagonalAccess<N>(tmpMain, tmpSuper, tmpSub, this.getStaticZero());
    }

    private final DecompositionStore<N> makeQ1(final DecompositionStore<N> aStoreToTransform, final boolean eye) {

        final DecompositionStore.HouseholderReference<N> tmpHouseholderReference = new DecompositionStore.HouseholderReference<N>(this.getInPlace(), BOOLEAN_TRUE);

        final int tmpRowDim = aStoreToTransform.getRowDim();
        final int tmpColDim = aStoreToTransform.getColDim();

        final boolean tmpUpper = this.isUpper();
        for (int ij = (tmpUpper && (tmpRowDim != tmpColDim)) ? tmpColDim - 1 : tmpColDim - 2; ij >= 0; ij--) {

            tmpHouseholderReference.row = tmpUpper ? ij : ij + 1;
            tmpHouseholderReference.col = ij;

            if (!tmpHouseholderReference.isZero()) {
                aStoreToTransform.transformLeft(tmpHouseholderReference, eye ? ij : 0);
            }
        }

        return aStoreToTransform;
    }

    private final DecompositionStore<N> makeQ2(final DecompositionStore<N> aStoreToTransform, final boolean eye) {

        final DecompositionStore.HouseholderReference<N> tmpHouseholderReference = new DecompositionStore.HouseholderReference<N>(this.getInPlace(), BOOLEAN_FALSE);

        final int tmpColDim = aStoreToTransform.getColDim();

        final boolean tmpUpper = this.isUpper();
        for (int ij = tmpUpper ? tmpColDim - 2 : tmpColDim - 1; ij >= 0; ij--) {

            tmpHouseholderReference.row = ij;
            tmpHouseholderReference.col = tmpUpper ? ij + 1 : ij;

            if (!tmpHouseholderReference.isZero()) {
                aStoreToTransform.transformLeft(tmpHouseholderReference, eye ? ij : 0);
            }
        }

        return aStoreToTransform;
    }

    /**
     * Will solve the equation system [aMtrxV][aMtrxD][X]=[aMtrxSimilar]<sup>T</sup> and overwrite the solution [X] to [aV].
     */
    private final void solve(final PhysicalStore<N> aMtrxV, final MatrixStore<N> aMtrxD, final DiagonalAccess<N> aMtrxSimilar) {

        final int tmpDim = aMtrxV.getRowDim();
        final int tmpLim = tmpDim - 1;

        double tmpSingular;
        for (int j = 0; j < tmpDim; j++) {
            tmpSingular = aMtrxD.doubleValue(j, j);
            if (TypeUtils.isZero(tmpSingular)) {
                for (int i = 0; i < tmpDim; i++) {
                    aMtrxV.set(i, j, PrimitiveMath.ZERO);
                }
            } else {
                for (int i = 0; i < tmpLim; i++) {
                    aMtrxV.set(i, j, ((aMtrxSimilar.doubleValue(i, i) * aMtrxV.doubleValue(i, j)) + (aMtrxSimilar.doubleValue(i, i + 1) * aMtrxV.doubleValue(i + 1, j))) / tmpSingular);
                }
                aMtrxV.set(tmpLim, j, (aMtrxSimilar.doubleValue(tmpLim, tmpLim) * aMtrxV.doubleValue(tmpLim, j)) / tmpSingular);
            }
        }
    }

    private final DecompositionStore<N> solve2(final PhysicalStore<N> aMtrxV, final MatrixStore<N> aMtrxD, final DiagonalAccess<N> aMtrxSimilar) {

        final int tmpDim = aMtrxV.getRowDim();
        final int tmpLim = tmpDim - 1;

        final DecompositionStore<N> retVal = this.makeZero(tmpDim, tmpDim);

        double tmpSingular;
        for (int j = 0; j < tmpDim; j++) {
            tmpSingular = aMtrxD.doubleValue(j, j);
            if (TypeUtils.isZero(tmpSingular)) {
                for (int i = 0; i < tmpDim; i++) {
                    retVal.set(i, j, aMtrxV.doubleValue(i, j));
                }
            } else {
                for (int i = 0; i < tmpLim; i++) {
                    retVal.set(i, j, ((aMtrxSimilar.doubleValue(i, i) * aMtrxV.doubleValue(i, j)) + (aMtrxSimilar.doubleValue(i, i + 1) * aMtrxV.doubleValue(i + 1, j))) / tmpSingular);
                }
                retVal.set(tmpLim, j, (aMtrxSimilar.doubleValue(tmpLim, tmpLim) * aMtrxV.doubleValue(tmpLim, j)) / tmpSingular);
            }
        }

        return retVal;
    }

    final DecompositionStore<N> doQ1(final DecompositionStore<N> aStoreToTransform) {
        return this.makeQ1(aStoreToTransform, BOOLEAN_FALSE);
    }

    final DecompositionStore<N> doQ2(final DecompositionStore<N> aStoreToTransform) {
        return this.makeQ2(aStoreToTransform, BOOLEAN_FALSE);
    }

    final DiagonalAccess<N> getDiagonalAccessD() {
        if (myDiagonalAccessD == null) {
            myDiagonalAccessD = this.makeDiagonalAccessD();
        }
        return myDiagonalAccessD;
    }

}
