/*
 * Copyright 2003-2011 Optimatika (www.optimatika.se)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.ojalgo.matrix.decomposition;

import java.util.ArrayList;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.ojalgo.access.Access2D;
import org.ojalgo.array.Array1D;
import org.ojalgo.concurrent.DaemonPoolExecutor;
import org.ojalgo.constant.PrimitiveMath;
import org.ojalgo.function.implementation.PrimitiveFunction;
import org.ojalgo.matrix.MatrixUtils;
import org.ojalgo.matrix.store.MatrixStore;
import org.ojalgo.matrix.store.PhysicalStore;
import org.ojalgo.matrix.store.PrimitiveDenseStore;
import org.ojalgo.matrix.transformation.Rotation;
import org.ojalgo.type.TypeUtils;
import org.ojalgo.type.context.NumberContext;

/**
 * Orginalet, sedan ett tag
 * 
 * Based on SVDnew2, but with transposing so that calculations are
 * always made on a matrix that "isAspectRationNormal".
 * 
 * Based on SVDnew5, but with Rotation replaced by the new alternative.
 * 
 * 
 *
 * @author apete
 */
abstract class SVDnew30<N extends Number & Comparable<N>> extends SingularValueDecomposition<N> {

    static final class Primitive extends SVDnew30<Double> {

        private static void doCase1(final double[] s, final double[] e, final int p, final int k, final ArrayList<Rotation<Double>> aRotQ2, final DecompositionStore<Double> aQ2) {

            double f = e[p - 2];
            e[p - 2] = PrimitiveMath.ZERO;

            double t;
            double cs;
            double sn;

            for (int j = p - 2; j >= k; j--) {

                t = Math.hypot(s[j], f);
                cs = s[j] / t;
                sn = f / t;

                s[j] = t;
                if (j != k) {
                    f = -sn * e[j - 1];
                    e[j - 1] = cs * e[j - 1];
                }

                if (aRotQ2 != null) {
                    aRotQ2.add(new Rotation.Primitive(p - 1, j, cs, sn));
                } else if (aQ2 != null) {
                    aQ2.rotateRight(p - 1, j, cs, sn);
                }
            }
        }

        private static void doCase2(final double[] s, final double[] e, final int p, final int k, final ArrayList<Rotation<Double>> aRotQ1, final DecompositionStore<Double> aQ1) {

            double f = e[k - 1];
            e[k - 1] = PrimitiveMath.ZERO;

            double t;
            double cs;
            double sn;

            for (int j = k; j < p; j++) {

                t = Math.hypot(s[j], f);
                cs = s[j] / t;
                sn = f / t;

                s[j] = t;
                f = -sn * e[j];
                e[j] = cs * e[j];

                if (aRotQ1 != null) {
                    aRotQ1.add(new Rotation.Primitive(k - 1, j, cs, sn));
                } else if (aQ1 != null) {
                    aQ1.rotateRight(k - 1, j, cs, sn);
                }
            }
        }

        private static void doCase3(final double[] s, final double[] e, final int p, final int k, final ArrayList<Rotation<Double>> aRotQ1, final ArrayList<Rotation<Double>> aRotQ2, final DecompositionStore<Double> aQ1, final DecompositionStore<Double> aQ2) {

            final int indPm1 = p - 1;
            final int indPm2 = p - 2;

            // Calculate the shift.
            final double scale = Math.max(Math.max(Math.max(Math.max(Math.abs(s[indPm1]), Math.abs(s[indPm2])), Math.abs(e[indPm2])), Math.abs(s[k])), Math.abs(e[k]));

            final double sPm1 = s[indPm1] / scale;
            final double sPm2 = s[indPm2] / scale;
            final double ePm2 = e[indPm2] / scale;
            final double sK = s[k] / scale;
            final double eK = e[k] / scale;

            final double b = (((sPm2 + sPm1) * (sPm2 - sPm1)) + (ePm2 * ePm2)) / PrimitiveMath.TWO;
            final double c = (sPm1 * ePm2) * (sPm1 * ePm2);

            double shift = PrimitiveMath.ZERO;
            if ((b != PrimitiveMath.ZERO) || (c != PrimitiveMath.ZERO)) {
                shift = Math.sqrt((b * b) + c);
                if (b < PrimitiveMath.ZERO) {
                    shift = -shift;
                }
                shift = c / (b + shift);
            }

            double f = ((sK + sPm1) * (sK - sPm1)) + shift;
            double g = sK * eK;

            double t;
            double cs;
            double sn;

            // Chase zeros.
            for (int j = k; j < indPm1; j++) {

                t = Math.hypot(f, g);
                cs = f / t;
                sn = g / t;
                if (j != k) {
                    e[j - 1] = t;
                }
                f = (cs * s[j]) + (sn * e[j]);
                e[j] = (cs * e[j]) - (sn * s[j]);
                g = sn * s[j + 1];
                s[j + 1] = cs * s[j + 1];

                if (aRotQ2 != null) {
                    aRotQ2.add(new Rotation.Primitive(j + 1, j, cs, sn));
                } else if (aQ2 != null) {
                    aQ2.rotateRight(j + 1, j, cs, sn);
                }

                t = Math.hypot(f, g);
                cs = f / t;
                sn = g / t;
                s[j] = t;
                f = (cs * e[j]) + (sn * s[j + 1]);
                s[j + 1] = (-sn * e[j]) + (cs * s[j + 1]);
                g = sn * e[j + 1];
                e[j + 1] = cs * e[j + 1];

                if (aRotQ1 != null) {
                    aRotQ1.add(new Rotation.Primitive(j + 1, j, cs, sn));
                } else if (aQ1 != null) {
                    aQ1.rotateRight(j + 1, j, cs, sn);
                }
            }

            e[indPm2] = f;
        }

        private static void doCase4(final double[] s, final int k, final ArrayList<Rotation<Double>> aRotQ1, final ArrayList<Rotation<Double>> aRotQ2, final DecompositionStore<Double> aQ1, final DecompositionStore<Double> aQ2) {

            final int tmpDiagDim = s.length;

            // Make the singular values positive.
            final double tmpSk = s[k];
            if (tmpSk < PrimitiveMath.ZERO) {
                s[k] = -tmpSk;
                if (aRotQ2 != null) {
                    aRotQ2.add(new Rotation.Primitive(k, k, Double.NaN, Double.NaN));
                } else if (aQ2 != null) {
                    aQ2.modifyColumn(0, k, PrimitiveFunction.NEGATE);
                }
            } else if (tmpSk == PrimitiveMath.ZERO) {
                s[k] = PrimitiveMath.ZERO; // To get rid of negative zeros
            }

            // Order the singular values.
            int tmpK = k;

            while (tmpK < (tmpDiagDim - 1)) {
                if (s[tmpK] >= s[tmpK + 1]) {
                    break;
                }
                final double t = s[tmpK];
                s[tmpK] = s[tmpK + 1];
                s[tmpK + 1] = t;

                if ((aRotQ1 != null) || (aRotQ2 != null)) {
                    final Rotation.Primitive tmpPrimitive = new Rotation.Primitive(tmpK + 1, tmpK, Double.NaN, Double.NaN);
                    aRotQ1.add(tmpPrimitive);
                    aRotQ2.add(tmpPrimitive);
                } else if ((aQ1 != null) || (aQ2 != null)) {
                    aQ1.exchangeColumns(tmpK + 1, tmpK);
                    aQ2.exchangeColumns(tmpK + 1, tmpK);
                }

                tmpK++;
            }
        }

        Primitive(final boolean aDirect) {
            super(PrimitiveDenseStore.FACTORY, new BidiagonalDecomposition.Primitive(), aDirect);
        }

        @Override
        Array1D<Double> compute(final Array1D<Double> mainDiagonal, final Array1D<Double> offDiagonal, final boolean singularValuesOnly) {

            final ArrayList<Rotation<Double>> tmpRotQ1 = (!singularValuesOnly && !this.isDirect()) ? this.getRotationsQ1() : null;
            final ArrayList<Rotation<Double>> tmpRotQ2 = (!singularValuesOnly && !this.isDirect()) ? this.getRotationsQ2() : null;

            final DecompositionStore<Double> tmpQ1 = (!singularValuesOnly && this.isDirect()) ? this.getBidiagonalQ1() : null;
            final DecompositionStore<Double> tmpQ2 = (!singularValuesOnly && this.isDirect()) ? this.getBidiagonalQ2() : null;

            final int tmpDiagDim = mainDiagonal.length;

            final double[] s = mainDiagonal.toRawCopy(); // s
            final double[] e = new double[tmpDiagDim]; // e
            final int tmpOffLength = offDiagonal.length;
            for (int i = 0; i < tmpOffLength; i++) {
                e[i] = offDiagonal.doubleValue(i);
            }

            // Main iteration loop for the singular values.
            int kase;
            int k;
            int p = tmpDiagDim;
            while (p > 0) {

                //
                // This section of the program inspects for negligible elements in the s and e arrays.
                // On completion the variables kase and k are set as follows:
                //
                // kase = 1     if s[p] and e[k-1] are negligible and k<p                           => deflate negligible s[p]
                // kase = 2     if s[k] is negligible and k<p                                       => split at negligible s[k]
                // kase = 3     if e[k-1] is negligible, k<p, and s(k)...s(p) are not negligible    => perform QR-step
                // kase = 4     if e[p-1] is negligible                                             => convergence.
                //

                kase = 0;
                k = 0;

                for (k = p - 2; k >= -1; k--) {
                    if (k == -1) {
                        break;
                    }
                    if (Math.abs(e[k]) <= (PrimitiveMath.MACHINE_DOUBLE_ERROR * (Math.abs(s[k]) + Math.abs(s[k + 1])))) {
                        e[k] = PrimitiveMath.ZERO;
                        break;
                    }
                }
                if (k == (p - 2)) {
                    kase = 4;
                } else {
                    int ks;
                    for (ks = p - 1; ks >= k; ks--) {
                        if (ks == k) {
                            break;
                        }
                        final double t = (ks != p ? Math.abs(e[ks]) : PrimitiveMath.ZERO) + (ks != (k + 1) ? Math.abs(e[ks - 1]) : PrimitiveMath.ZERO);
                        if (Math.abs(s[ks]) <= (PrimitiveMath.MACHINE_DOUBLE_ERROR * t)) {
                            s[ks] = PrimitiveMath.ZERO;
                            break;
                        }
                    }
                    if (ks == k) {
                        kase = 3;
                    } else if (ks == (p - 1)) {
                        kase = 1;
                    } else {
                        kase = 2;
                        k = ks;
                    }
                }
                k++;

                switch (kase) { // Perform the task indicated by kase.

                case 1: // Deflate negligible s[p]

                    Primitive.doCase1(s, e, p, k, tmpRotQ2, tmpQ2);
                    break;

                case 2: // Split at negligible s[k]

                    Primitive.doCase2(s, e, p, k, tmpRotQ1, tmpQ1);
                    break;

                case 3: // Perform QR-step.

                    Primitive.doCase3(s, e, p, k, tmpRotQ1, tmpRotQ2, tmpQ1, tmpQ2);
                    break;

                case 4: // Convergence

                    Primitive.doCase4(s, k, tmpRotQ1, tmpRotQ2, tmpQ1, tmpQ2);
                    p--;
                    break;

                default:

                    throw new IllegalStateException();

                } // switch
            } // while

            if (tmpQ1 != null) {
                this.setQ1(tmpQ1);
            }
            if (tmpQ2 != null) {
                this.setQ2(tmpQ2);
            }

            return Array1D.wrapPrimitive(s);
        }
    }

    private transient Future<MatrixStore<N>> myFutureQ1;
    private transient Future<MatrixStore<N>> myFutureQ2;

    private final ArrayList<Rotation<N>> myRotationsQ1 = new ArrayList<Rotation<N>>();
    private final ArrayList<Rotation<N>> myRotationsQ2 = new ArrayList<Rotation<N>>();

    private final boolean myDirect;

    @SuppressWarnings("unused")
    private SVDnew30(final DecompositionStore.Factory<N, ? extends DecompositionStore<N>> aFactory, final BidiagonalDecomposition<N> aBidiagonal) {
        this(aFactory, aBidiagonal, BOOLEAN_FALSE);
    }

    protected SVDnew30(final DecompositionStore.Factory<N, ? extends DecompositionStore<N>> aFactory, final BidiagonalDecomposition<N> aBidiagonal, final boolean aDirect) {

        super(aFactory, aBidiagonal);

        myDirect = aDirect;
    }

    public boolean equals(final MatrixStore<N> aMtrx, final NumberContext aCntxt) {
        return MatrixUtils.equals(aMtrx, this, TypeUtils.EQUALS_NUMBER_CONTEXT);
    }

    public boolean isFullSize() {
        return BOOLEAN_FALSE;
    }

    public boolean isOrdered() {
        return BOOLEAN_TRUE;
    }

    public boolean isSolvable() {
        return this.isComputed();
    }

    @Override
    public void reset() {

        super.reset();

        myFutureQ1 = null;
        myFutureQ2 = null;

        myRotationsQ1.clear();
        myRotationsQ2.clear();
    }

    @Override
    public final MatrixStore<N> solve(final MatrixStore<N> aRHS) {
        return this.getInverse().multiplyRight(aRHS);
    }

    private final void setupQ1andQ2calculations() {

        //    BasicLogger.logDebug("Q1 Rotation# {}", myRotationsQ1.size());
        //      BasicLogger.logDebug("Q2 Rotation# {}", myRotationsQ2.size());

        myFutureQ1 = DaemonPoolExecutor.INSTANCE.submit(new Callable<MatrixStore<N>>() {

            public MatrixStore<N> call() throws Exception {

                final PhysicalStore<N> retVal = SVDnew30.this.getBidiagonalQ1();

                for (final Rotation<N> tmpRot : myRotationsQ1) {
                    retVal.transformRight(tmpRot);
                }

                return retVal;
            }
        });

        myFutureQ2 = DaemonPoolExecutor.INSTANCE.submit(new Callable<MatrixStore<N>>() {

            public MatrixStore<N> call() throws Exception {

                final PhysicalStore<N> retVal = SVDnew30.this.getBidiagonalQ2();

                for (final Rotation<N> tmpRot : myRotationsQ2) {
                    retVal.transformRight(tmpRot);
                }

                return retVal;
            }

        });

    }

    @Override
    protected boolean doCompute(final Access2D<?> aMtrx, final boolean singularValuesOnly) {

        this.computeBidiagonal(aMtrx);

        final DiagonalAccess<N> tmpBidiagonal = this.getBidiagonalAccessD();

        this.setSingularValues(this.compute(tmpBidiagonal.mainDiagonal, tmpBidiagonal.superdiagonal, singularValuesOnly));

        return this.computed(BOOLEAN_TRUE);
    }

    @Override
    protected final MatrixStore<N> makeD() {
        return this.wrap(new DiagonalAccess<Double>(this.getSingularValues(), null, null, PrimitiveMath.ZERO));
    }

    @Override
    protected final MatrixStore<N> makeQ1() {

        if ((myFutureQ1 == null) || (myFutureQ2 == null)) {
            this.setupQ1andQ2calculations();
        }

        try {
            return myFutureQ1.get();
        } catch (final InterruptedException anException) {
            anException.printStackTrace();
            return null;
        } catch (final ExecutionException anException) {
            anException.printStackTrace();
            return null;
        }
    }

    @Override
    protected final MatrixStore<N> makeQ2() {

        if ((myFutureQ1 == null) || (myFutureQ2 == null)) {
            this.setupQ1andQ2calculations();
        }

        try {
            return myFutureQ2.get();
        } catch (final InterruptedException anException) {
            anException.printStackTrace();
            return null;
        } catch (final ExecutionException anException) {
            anException.printStackTrace();
            return null;
        }
    }

    @Override
    protected Array1D<Double> makeSingularValues() {
        throw new IllegalStateException("Should never have to be called!");
    }

    abstract Array1D<Double> compute(Array1D<N> mainDiagonal, Array1D<N> offDiagonal, final boolean singularValuesOnly);

    final ArrayList<Rotation<N>> getRotationsQ1() {
        return myRotationsQ1;
    }

    final ArrayList<Rotation<N>> getRotationsQ2() {
        return myRotationsQ2;
    }

    final boolean isDirect() {
        return myDirect;
    }

}
