/*
 * Copyright 2003-2011 Optimatika (www.optimatika.se)
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.ojalgo.matrix.decomposition;

import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Future;

import org.ojalgo.access.Access2D;
import org.ojalgo.array.Array1D;
import org.ojalgo.concurrent.DaemonPoolExecutor;
import org.ojalgo.constant.PrimitiveMath;
import org.ojalgo.function.BinaryFunction;
import org.ojalgo.function.PreconfiguredSecond;
import org.ojalgo.function.aggregator.AggregatorFunction;
import org.ojalgo.function.aggregator.ComplexAggregator;
import org.ojalgo.matrix.MatrixUtils;
import org.ojalgo.matrix.decomposition.DecompositionStore.ExchangeColumnsCaller;
import org.ojalgo.matrix.decomposition.DecompositionStore.RotateRightCaller;
import org.ojalgo.matrix.store.MatrixStore;
import org.ojalgo.matrix.store.PhysicalStore;
import org.ojalgo.matrix.store.PrimitiveDenseStore;
import org.ojalgo.scalar.ComplexNumber;
import org.ojalgo.type.TypeUtils;
import org.ojalgo.type.context.NumberContext;

abstract class SymmetricEvD1<N extends Number> extends EigenvalueDecomposition<N> {

    /** Eigenvalues and eigenvectors of a real matrix. 
     <P>
     If A is symmetric, then A = V*D*V' where the eigenvalue matrix D is
     diagonal and the eigenvector matrix V is orthogonal.
     I.e. A = V.times(D.times(V.transpose())) and 
     V.times(V.transpose()) equals the identity matrix.
     <P>
     If A is not symmetric, then the eigenvalue matrix D is block diagonal
     with the real eigenvalues in 1-by-1 blocks and any complex eigenvalues,
     lambda + i*mu, in 2-by-2 blocks, [lambda, mu; -mu, lambda].  The
     columns of V represent the eigenvectors in the sense that A*V = V*D,
     i.e. A.times(V) equals V.times(D).  The matrix V may be badly
     conditioned, or even singular, so the validity of the equation
     A = V*D*inverse(V) depends upon V.cond().
     **/
    static final class Primitive extends SymmetricEvD1<Double> {

        Primitive() {
            super(PrimitiveDenseStore.FACTORY, new TridiagonalDecomposition.Primitive(BOOLEAN_TRUE));
        }

        @Override
        protected Array1D<Double> toDiagonal(final DiagonalAccess<Double> aTridiagonal, final DecompositionStore<Double> aV) {

            final double[] tmpMainDiagonal = aTridiagonal.mainDiagonal.toRawCopy();

            final int tmpDim = tmpMainDiagonal.length;
            final int tmpLim = tmpDim - 1;

            final double[] tmpOffDiagonal = new double[tmpDim];
            final Array1D<Double> tmpSuperdiagonal = aTridiagonal.superdiagonal;
            for (int i = 0; i < tmpLim; i++) {
                tmpOffDiagonal[i] = tmpSuperdiagonal.doubleValue(i);
            }

            //                BasicLogger.logDebug("BEGIN diagonalize");
            //                BasicLogger.logDebug("Main D: {}", Arrays.toString(tmpMainDiagonal));
            //                BasicLogger.logDebug("Seco D: {}", Arrays.toString(tmpOffDiagonal));
            //                BasicLogger.logDebug("V", aV);
            //                BasicLogger.logDebug();

            double tmpShift = PrimitiveMath.ZERO;
            double tmpShiftIncr;

            double tmpMagnitude = PrimitiveMath.ZERO;
            double tmpLocalEpsilon;

            int m;
            // Main loop
            for (int l = 0; l < tmpDim; l++) {

                //  EigenvalueDecomposition.log("Loop l=" + l, tmpMainDiagonal, tmpOffDiagonal);

                // Find small subdiagonal element
                tmpMagnitude = Math.max(tmpMagnitude, Math.abs(tmpMainDiagonal[l]) + Math.abs(tmpOffDiagonal[l]));
                tmpLocalEpsilon = PrimitiveMath.MACHINE_DOUBLE_ERROR * tmpMagnitude;

                m = l;
                while (m < tmpDim) {
                    if (Math.abs(tmpOffDiagonal[m]) <= tmpLocalEpsilon) {
                        break;
                    }
                    m++;
                }

                // If m == l, aMainDiagonal[l] is an eigenvalue, otherwise, iterate.
                if (m > l) {

                    do {

                        final double tmp1Ml0 = tmpMainDiagonal[l];
                        final double tmp1Ml1 = tmpMainDiagonal[l + 1];
                        final double tmp1Sl0 = tmpOffDiagonal[l];

                        // Compute implicit shift

                        double p = (tmp1Ml1 - tmp1Ml0) / (tmp1Sl0 + tmp1Sl0);
                        double r = Math.hypot(p, PrimitiveMath.ONE);
                        if (p < 0) {
                            r = -r;
                        }

                        final double tmp2Ml0 = tmpMainDiagonal[l] = tmp1Sl0 / (p + r);
                        final double tmp2Ml1 = tmpMainDiagonal[l + 1] = tmp1Sl0 * (p + r);
                        final double tmp2Sl1 = tmpOffDiagonal[l + 1];

                        tmpShiftIncr = tmp1Ml0 - tmp2Ml0;
                        for (int i = l + 2; i < tmpDim; i++) {
                            tmpMainDiagonal[i] -= tmpShiftIncr;
                        }
                        tmpShift += tmpShiftIncr;

                        //     EigenvalueDecomposition.log("New shift =" + tmpShift, tmpMainDiagonal, tmpOffDiagonal);

                        // Implicit QL transformation

                        double tmpRotCos = PrimitiveMath.ONE;
                        double tmpRotSin = PrimitiveMath.ZERO;

                        double tmpRotCos2 = tmpRotCos;
                        double tmpRotSin2 = PrimitiveMath.ZERO;

                        double tmpRotCos3 = tmpRotCos;

                        p = tmpMainDiagonal[m]; // Initiate p
                        for (int i = m - 1; i >= l; i--) {

                            final double tmp1Mi0 = tmpMainDiagonal[i];
                            final double tmp1Si0 = tmpOffDiagonal[i];

                            r = Math.hypot(p, tmp1Si0);

                            tmpRotCos3 = tmpRotCos2;

                            tmpRotCos2 = tmpRotCos;
                            tmpRotSin2 = tmpRotSin;

                            tmpRotCos = p / r;
                            tmpRotSin = tmp1Si0 / r;

                            tmpMainDiagonal[i + 1] = (tmpRotCos2 * p) + (tmpRotSin * ((tmpRotCos * tmpRotCos2 * tmp1Si0) + (tmpRotSin * tmp1Mi0)));
                            tmpOffDiagonal[i + 1] = tmpRotSin2 * r;

                            p = (tmpRotCos * tmp1Mi0) - (tmpRotSin * tmpRotCos2 * tmp1Si0); // Next p

                            // Accumulate transformation - rotate the eigenvector matrix
                            //aV.transformRight(new Rotation.Primitive(i, i + 1, tmpRotCos, tmpRotSin));

                            if (aV != null) {
                                //myFutureV = EXEC.submit(new CallRotateRight(aV, i, i + 1, tmpRotCos, tmpRotSin));
                                this.add(new RotateRightCaller<Double>(aV, i, i + 1, tmpRotCos, tmpRotSin));
                                //final Rotation.Primitive tmpRotation = new Rotation.Primitive(i, i + 1, tmpRotCos, tmpRotSin);
                                //BasicLogger.logDebug("Created: " + tmpRotation);
                                //this.addRotation(tmpRotation);
                            }

                            //          EigenvalueDecomposition.log("QL step done i=" + i, tmpMainDiagonal, tmpOffDiagonal);

                        }

                        p = (-tmpRotSin * tmpRotSin2 * tmpRotCos3 * tmp2Sl1 * tmpOffDiagonal[l]) / tmp2Ml1; // Final p

                        tmpMainDiagonal[l] = tmpRotCos * p;
                        tmpOffDiagonal[l] = tmpRotSin * p;

                    } while (Math.abs(tmpOffDiagonal[l]) > tmpLocalEpsilon); // Check for convergence
                } // End if (m > l)

                final double tmpEigenvalue = tmpMainDiagonal[l] + tmpShift;
                if (TypeUtils.isZero(tmpEigenvalue)) {
                    tmpMainDiagonal[l] = PrimitiveMath.ZERO;
                } else {
                    tmpMainDiagonal[l] = tmpEigenvalue;
                }
                tmpOffDiagonal[l] = PrimitiveMath.ZERO;
            } // End main loop - l

            //                BasicLogger.logDebug("END diagonalize");
            //                BasicLogger.logDebug("Main D: {}", Arrays.toString(tmpMainDiagonal));
            //                BasicLogger.logDebug("Seco D: {}", Arrays.toString(tmpOffDiagonal));
            //                BasicLogger.logDebug("V", aV);
            //                BasicLogger.logDebug();

            return Array1D.wrapPrimitive(tmpMainDiagonal);
        }

    }

    private static final DaemonPoolExecutor EXECUTOR_V = DaemonPoolExecutor.makeSingle();
    private Array1D<N> myDiagonalValues;
    private Future<DecompositionStore<N>> myFutureV = null;
    private final TridiagonalDecomposition<N> myTridiagonal;

    @SuppressWarnings("unused")
    private SymmetricEvD1(final DecompositionStore.Factory<N, ? extends DecompositionStore<N>> aFactory) {
        this(aFactory, null);
    }

    protected SymmetricEvD1(final DecompositionStore.Factory<N, ? extends DecompositionStore<N>> aFactory, final TridiagonalDecomposition<N> aTridiagonal) {

        super(aFactory);

        myTridiagonal = aTridiagonal;
    }

    public final boolean compute(final Access2D<?> aMtrx, final boolean eigenvaluesOnly) {
        return this.compute(aMtrx, BOOLEAN_TRUE, eigenvaluesOnly);
    }

    public final boolean equals(final MatrixStore<N> aStore, final NumberContext aCntxt) {
        return MatrixUtils.equals(aStore, this, aCntxt);
    }

    public final ComplexNumber getDeterminant() {

        final AggregatorFunction<ComplexNumber> tmpVisitor = ComplexAggregator.getCollection().product();

        this.getEigenvalues().visitAll(tmpVisitor);

        return tmpVisitor.getNumber();
    }

    public final ComplexNumber getTrace() {

        final AggregatorFunction<ComplexNumber> tmpVisitor = ComplexAggregator.getCollection().sum();

        this.getEigenvalues().visitAll(tmpVisitor);

        return tmpVisitor.getNumber();
    }

    public final boolean isFullSize() {
        return BOOLEAN_TRUE;
    }

    public final boolean isOrdered() {
        return BOOLEAN_TRUE;
    }

    public final boolean isSolvable() {
        return this.isComputed() && this.isSymmetric();
    }

    public final boolean isSymmetric() {
        return BOOLEAN_TRUE;
    }

    @Override
    public void reset() {

        super.reset();

        myTridiagonal.reset();
        myFutureV = null;
    }

    @Override
    public final MatrixStore<N> solve(final MatrixStore<N> aRHS) {
        return this.getInverse().multiplyRight(aRHS);
    }

    protected final void add(final Callable<DecompositionStore<N>> aCallable) {
        myFutureV = EXECUTOR_V.submit(aCallable);
    }

    @Override
    protected final boolean doNonsymmetric(final Access2D<?> aMtrx, final boolean eigenvaluesOnly) {
        throw new UnsupportedOperationException();
    }

    @Override
    protected final boolean doSymmetric(final Access2D<?> aMtrx, final boolean eigenvaluesOnly) {

        final int tmpDim = aMtrx.getRowDim();

        myTridiagonal.compute(aMtrx);

        final DiagonalAccess<N> tmpTridiagonal = myTridiagonal.getDiagonalAccessD();

        final DecompositionStore<N> tmpV = eigenvaluesOnly ? null : myTridiagonal.makeQ(BOOLEAN_TRUE);

        final Array1D<N> tmpDiagonal = myDiagonalValues = this.toDiagonal(tmpTridiagonal, tmpV);

        final int tmpLim = tmpDim - 1;

        for (int j1 = 0; j1 < tmpLim; j1++) {

            double tmpValue = tmpDiagonal.doubleValue(j1);
            int j2 = j1;

            for (int ij1 = j1 + 1; ij1 < tmpDim; ij1++) {

                if (Math.abs(tmpDiagonal.doubleValue(ij1)) > Math.abs(tmpValue)) {

                    tmpValue = tmpDiagonal.doubleValue(ij1);
                    j2 = ij1;
                }
            }

            if (j2 != j1) {
                tmpDiagonal.set(j2, tmpDiagonal.doubleValue(j1));
                tmpDiagonal.set(j1, tmpValue);
                //eigenvectors.exchangeColumns(j1, j2);
                if (!eigenvaluesOnly) {
                    //myFutureV = EXEC.submit(new CallExchangeColumns(tmpV, j1, j2));
                    this.add(new ExchangeColumnsCaller<N>(tmpV, j1, j2));
                    //this.addRotation(this.makeRotation(j1, j2, null, null));
                }

            }
        }

        return this.computed(BOOLEAN_TRUE);
    }

    @Override
    protected boolean isSmallEnoughToFitInCache() {
        return myTridiagonal.isSmallEnoughToFitInCache();
    }

    @Override
    protected MatrixStore<N> makeD() {

        final int tmpDim = myDiagonalValues.length;

        final PhysicalStore<N> retVal = this.makeZero(tmpDim, tmpDim);

        for (int ij = 0; ij < tmpDim; ij++) {
            retVal.set(ij, ij, myDiagonalValues.doubleValue(ij));
        }

        return retVal;
    }

    @Override
    protected Array1D<ComplexNumber> makeEigenvalues() {

        final int tmpDim = myDiagonalValues.length;

        final Array1D<ComplexNumber> retVal = Array1D.makeComplex(tmpDim);

        for (int ij = 0; ij < tmpDim; ij++) {
            retVal.set(ij, new ComplexNumber(myDiagonalValues.doubleValue(ij)));
        }

        return retVal;
    }

    @Override
    protected final MatrixStore<N> makeInverse() {

        final MatrixStore<N> tmpV = this.getV();
        final MatrixStore<N> tmpD = this.getD();

        final int tmpDim = tmpD.getRowDim();

        final PhysicalStore<N> tmpMtrx = tmpV.transpose();

        final N tmpZero = this.getStaticZero();
        final BinaryFunction<N> tmpDivide = this.getFunctionSet().divide();

        for (int i = 0; i < tmpDim; i++) {
            if (tmpD.isZero(i, i)) {
                tmpMtrx.fillRow(i, 0, tmpZero);
            } else {
                tmpMtrx.modifyRow(i, 0, new PreconfiguredSecond<N>(tmpDivide, tmpD.get(i, i)));
            }
        }

        return tmpMtrx.multiplyLeft(tmpV);
    }

    @Override
    protected MatrixStore<N> makeV() {
        try {
            if (myFutureV != null) {
                return myFutureV.get();
            } else {
                return myTridiagonal.makeQ(BOOLEAN_TRUE);
            }
        } catch (final InterruptedException anException) {
            anException.printStackTrace();
            return null;
        } catch (final ExecutionException anException) {
            anException.printStackTrace();
            return null;
        }
    }

    protected abstract Array1D<N> toDiagonal(final DiagonalAccess<N> aTridiagonalSymmetric, final DecompositionStore<N> aV);

}
