/* 
 * Copyright 2003-2011 Optimatika (www.optimatika.se)
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.ojalgo.optimisation;

import static org.ojalgo.constant.BigMath.*;

import java.math.BigDecimal;
import java.math.MathContext;

import org.ojalgo.function.aggregator.AggregatorFunction;
import org.ojalgo.function.aggregator.BigAggregator;
import org.ojalgo.type.context.NumberContext;

/**
 * Variable
 * 
 * @author apete
 */
public class Variable extends ModelComponent<Variable> {

    public static Variable make(final String aName) {
        return new Variable(aName);
    }

    public static Variable makeBinary(final String aName) {
        return Variable.make(aName).binary();
    }

    private transient BigDecimal myAdjustmentFactor;
    private boolean myInteger = false;
    private BigDecimal myValue = null;

    public Variable(final String aName) {
        super(aName);
    }

    public final Variable binary() {
        return this.lower(ZERO).upper(ONE).integer(true);
    }

    public Variable copy() {

        final Variable retVal = new Variable(this.getName());

        retVal.setContributionWeight(this.getContributionWeight());
        retVal.setLowerLimit(this.getLowerLimit());
        retVal.setUpperLimit(this.getUpperLimit());
        retVal.setValue(myValue);
        retVal.setInteger(myInteger);

        return retVal;
    }

    public final BigDecimal getAdjustedUnit() {
        return this.getAdjustmentFactor();
    }

    @Override
    public BigDecimal getAdjustmentFactor() {

        if (myAdjustmentFactor == null) {

            final AggregatorFunction<BigDecimal> tmpLargestAggr = BigAggregator.getCollection().largest();
            final AggregatorFunction<BigDecimal> tmpSmallestAggr = BigAggregator.getCollection().smallest();

            tmpLargestAggr.invoke(ONE);
            tmpSmallestAggr.invoke(ONE);

            if (this.getLowerLimit() != null) {
                tmpLargestAggr.invoke(this.getLowerLimit());
                tmpSmallestAggr.invoke(this.getLowerLimit());
            }

            if (this.getUpperLimit() != null) {
                tmpLargestAggr.invoke(this.getUpperLimit());
                tmpSmallestAggr.invoke(this.getUpperLimit());
            }

            final int tmpExponent = OptimisationUtils.getAdjustmentFactorExponent(tmpLargestAggr, tmpSmallestAggr);

            myAdjustmentFactor = BigDecimal.TEN.pow(tmpExponent, MathContext.DECIMAL32);
        }

        return myAdjustmentFactor;
    }

    public BigDecimal getLowerSlack() {

        BigDecimal retVal = null;

        if (this.getLowerLimit() != null) {

            if (myValue != null) {

                retVal = this.getLowerLimit().subtract(myValue);

            } else {

                retVal = this.getLowerLimit();
            }
        }

        if ((retVal != null) && this.isInteger()) {
            retVal = retVal.setScale(0, BigDecimal.ROUND_CEILING);
        }

        return retVal;
    }

    public BigDecimal getUpperSlack() {

        BigDecimal retVal = null;

        if (this.getUpperLimit() != null) {

            if (myValue != null) {

                retVal = this.getUpperLimit().subtract(myValue);

            } else {

                retVal = this.getUpperLimit();
            }
        }

        if ((retVal != null) && this.isInteger()) {
            retVal = retVal.setScale(0, BigDecimal.ROUND_FLOOR);
        }

        return retVal;
    }

    public BigDecimal getValue() {
        return myValue;
    }

    public Variable integer(final boolean aBool) {
        this.setInteger(aBool);
        return this;
    }

    public boolean isBinary() {

        boolean retVal = this.isInteger();

        retVal &= this.isLowerConstraint() && this.getLowerLimit().equals(ZERO);

        retVal &= this.isUpperConstraint() && this.getUpperLimit().equals(ONE);

        return retVal;
    }

    public boolean isInteger() {
        return myInteger;
    }

    public boolean isValueSet() {
        return myValue != null;
    }

    public BigDecimal quantifyContribution() {

        BigDecimal retVal = ZERO;

        final BigDecimal tmpContributionWeight = this.getContributionWeight();
        if ((tmpContributionWeight != null) && (myValue != null)) {
            retVal = tmpContributionWeight.multiply(myValue);
        }

        return retVal;
    }

    public void setInteger(final boolean aBool) {
        myInteger = aBool;
    }

    public void setValue(final BigDecimal aValue) {
        myValue = aValue;
    }

    public boolean validateState(final NumberContext aContext) {
        try {
            return super.validate(myValue, aContext);
        } catch (final ModelValidationException anException) {
            System.err.println(anException.toString());
            return false;
        }
    }

    @Override
    protected void appendMiddlePart(final StringBuilder aStringBuilder) {

        aStringBuilder.append(this.getName());

        if (myValue != null) {
            aStringBuilder.append(": ");
            aStringBuilder.append(OptimisationUtils.DISPLAY.enforce(myValue).toPlainString());
        }

        if (this.isObjective()) {
            aStringBuilder.append(" (");
            aStringBuilder.append(OptimisationUtils.DISPLAY.enforce(this.getContributionWeight()).toPlainString());
            aStringBuilder.append(")");
        }
    }
}
